/* Copyright (C) 2007, 2008, 2009, 2010 Stefan Frenz
 * 
 * This file is part of SJC, the Small Java Compiler written by Stefan Frenz.
 * 
 * SJC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * SJC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with SJC. If not, see <http://www.gnu.org/licenses/>.
 */

package frontend.sjava;

import backend.Instruction;
import compbase.*;

/**
 * ExDeArray: array dereferenzation
 * 
 * @author S. Frenz
 * @version 100319 allowed unlimited indices for struct inline arrays
 *  version 100312 adopted changed TypeRef and added support for flash objects
 *  version 091112 added support for explicitTypeConversion
 *  version 091021 adopted changed modifier declarations
 *  version 091005 adopted changed Expr with support for preferredType in resolving
 *  version 091001 adopted changed memory interface
 *  version 090724 adopted changed Expr
 *  version 090718 adopted changed Expr
 *  version 090702 fixed invalid address register recycling in genOutputAddr
 *  version 090619 adopted changed Architecture
 *  version 090207 added copyright notice
 *  version 090206 adopted changed Expr.genOutputAssignTo
 *  version 081021 adopted changed Architecture.genComp/genCondJmp
 *  version 080525 adopted changed genCondJmp signature
 *  version 080122 optimized push-pop-sequence for type-checking in assign
 *  version 080119 added support for type-checking in assign
 *  version 070909 optimized signature of Expr.resolve
 *  version 070727 replaced exSubResolve by resolve as there is nothing done anymore
 *  version 070628 adopted changes for "Expr extends TypeRef"
 *  version 070528 optimized unit import check
 *  version 070526 added support for inline arrays in instances
 *  version 070331 added support for additional indirect scalars in SArray
 *  version 070303 added support for movable indirect scalars
 *  version 070208 fixed object array offset
 *  version 070114 fixed SSA-relevent register duplication, reduced access level where possible
 *  version 070113 adopted change of genCheckNull to genCompPtrToNull
 *  version 070111 adapted change in printPos and compErr
 *  version 070106 initial version
 */

public class ExDeArray extends Expr {
  protected Expr le, ind;
  private UnitList runtimeClass;
  
  protected ExDeArray(int fid, int il, int ic) {
    super(fid, il, ic);
  }
  
  public boolean resolve(Unit unitContext, Mthd mthdContext, boolean checkRead, boolean insideLoop,
      TypeRef preferredElemType, Context ctx) {
    TypeRef preferredObjType=null;
    
    if (preferredElemType!=null) {
      preferredObjType=new TypeRef(fileID, line, col);
      preferredObjType.getTypeOf(preferredElemType);
      preferredObjType.arrDim++;
    }
    if (!le.resolve(unitContext, mthdContext, true, insideLoop, preferredObjType, ctx)) return false;
    if (!ind.resolve(unitContext, mthdContext, true, insideLoop, ctx.intType, ctx)) {
      ctx.out.print(" in array-deref");
      return false;
    }
    if (le.arrDim<1) {
      printPos(ctx, "left side is not an array");
      return false;
    }
    if ((mthdContext.modifier&Modifier.MC_EXPCONV)==0) switch (ind.getBaseTypeConvertible()) {
      case StdTypes.T_BYTE: case StdTypes.T_SHRT:
        if ((ind=ExEnc.getConvertedResolvedExpr(ind, ctx.intType, unitContext, ctx))==null) return false;
    }
    if (!ind.isIntType()) {
      printPos(ctx, "need int-type in array-deref");
      return false;
    }
    getTypeOf(le); //get a copy, we will modify arrDim
    arrDim--;
    if (isIntfType()) {
      printPos(ctx, "interface-arrays not supported");
      return false;
    }
    //import runtimeClass if necessary
    if (ctx.dynaMem && (ctx.runtimeBound || ctx.runtimeNull || ctx.doArrayStoreCheck))
      runtimeClass=unitContext.getRefUnit(ctx.rteDynamicRuntime, true);
    //everything ok
    return true;
  }
  
  public boolean canGenAddr(Unit unitContext, boolean allowSpecialWriteAccess, boolean insideLoop, Context ctx) {
    if (ctx.dynaMem && ctx.doArrayStoreCheck && runtimeClass==null)
      runtimeClass=unitContext.getRefUnit(ctx.rteDynamicRuntime, true);
    return true;
  }
  
  public void genOutputAddr(int reg, Context ctx) {
    genOutputAddr(0, reg, ctx);
  }
  
  public void genOutputPrepareAssignTo(int destReg, int newValueReg, Expr newValueEx, Context ctx) {
    int lePtr, restore1, restore2;
    
    if (ctx.doArrayStoreCheck && newValueEx.isObjType()) {
      restore1=ctx.arch.prepareFreeReg(destReg, newValueReg, 0, StdTypes.T_PTR);
      lePtr=ctx.arch.allocReg();
      genOutputAddr(lePtr, destReg, ctx);
      newValueEx.genOutputVal(newValueReg, ctx);
      restore2=ctx.arch.ensureFreeRegs(lePtr, newValueReg, 0, 0);
      ctx.arch.genPush(newValueReg, StdTypes.T_PTR);
      if (ctx.dynaMem) ctx.arch.genSaveUnitContext();
      ctx.arch.genPush(lePtr, StdTypes.T_PTR);
      ctx.arch.genPush(newValueReg, StdTypes.T_PTR);
      if (ctx.dynaMem) {
        ctx.arch.genLoadUnitContext(ctx.arch.regClss, runtimeClass.relOff);
        ctx.arch.genCall(ctx.rteDRCheckArrayStoreMd.relOff, 0, true,
            ctx.rteDRCheckArrayStoreMd.parSize);
        ctx.arch.genRestUnitContext();
      }
      else ctx.arch.genCallConst(ctx.rteDRCheckArrayStoreMd, ctx.rteDRCheckArrayStoreMd.parSize);
      ctx.arch.genPop(newValueReg, StdTypes.T_PTR);
      ctx.arch.deallocRestoreReg(0, 0, restore2);
      ctx.arch.deallocRestoreReg(lePtr, 0, restore1);
    }
    else super.genOutputPrepareAssignTo(destReg, newValueReg, newValueEx, ctx);
  }
  
  private void genOutputAddr(int lePtr, int addrReg, Context ctx) {
    int obj, objRestore=0, pos, restore, index, regType, size, lenPos=ctx.rteSArrayLength, condHnd;
    Unit container=ctx.rteSArray;
    AccVar inlName;
    Vrbl inlDecl;
    Instruction excCheckDone;
    boolean flash=false;
    
    if (lePtr==0) { //left side is not needed after outputAddr
      objRestore=ctx.arch.prepareFreeReg(0, 0, addrReg, StdTypes.T_PTR);
      obj=ctx.arch.allocReg();
    }
    else obj=lePtr; //remember pointer to array for caller
    le.genOutputVal(obj, ctx);
    if (le.typeSpecial==TypeRef.S_FLASHINLARR || le.typeSpecial==TypeRef.S_FLASHREF) flash=true;
    switch (le.typeSpecial) {
      case TypeRef.S_FLASHINLARR: //flash instance inline array
      case TypeRef.S_INSTINLARR: //instance inline array
        if (le instanceof ExVar) inlName=((ExVar)le).dest;
        else if (le instanceof ExDeRef && ((ExDeRef)le).ri instanceof ExVar)
          inlName=((ExVar)((ExDeRef)le).ri).dest;
        else {
          compErr(ctx, "could not resolve declaration of inline array");
          return;
        }
        if (inlName instanceof Vrbl) inlDecl=(Vrbl)inlName;
        else {
          compErr(ctx, "could not resolve variable of inline array");
          return;
        }
        lenPos=inlDecl.relOff;
        container=inlDecl.owner;
        //no break
      case TypeRef.S_FLASHREF: //flash instance array (both normal and inline)
      case TypeRef.S_NOSPECIAL: //instance array (both normal and inline)
        if (ctx.runtimeNull) {
          excCheckDone=ctx.arch.getUnlinkedInstruction();
          condHnd=ctx.arch.genCompPtrToNull(obj, Ops.C_NE);
          ctx.arch.genCondJmp(excCheckDone, condHnd);
          if (ctx.dynaMem) {
            ctx.arch.genSaveUnitContext();
            ctx.arch.genLoadUnitContext(ctx.arch.regClss, runtimeClass.relOff);
            ctx.arch.genCall(ctx.rteDRNullExcMd.relOff, 0, true,
                ctx.rteDRNullExcMd.parSize);
            ctx.arch.genRestUnitContext();
          }
          else ctx.arch.genCallConst(ctx.rteDRNullExcMd, ctx.rteDRNullExcMd.parSize);
          ctx.arch.appendInstruction(excCheckDone);
        }
        restore=ctx.arch.prepareFreeReg(obj, addrReg, 0, StdTypes.T_INT);
        index=ctx.arch.allocReg();
        ind.genOutputVal(index, ctx);
        if (ctx.doBoundCheck) {
          excCheckDone=ctx.arch.getUnlinkedInstruction();
          ctx.arch.genCheckBounds(obj, index, lenPos, excCheckDone);
          if (ctx.runtimeBound) {
            if (ctx.dynaMem) {
              ctx.arch.genSaveUnitContext();
              ctx.arch.genPush(obj, StdTypes.T_PTR);
              ctx.arch.genPush(index, StdTypes.T_INT);
              ctx.arch.genLoadUnitContext(ctx.arch.regClss, runtimeClass.relOff);
              ctx.arch.genCall(ctx.rteDRBoundExcMd.relOff, 0, true,
                  ctx.rteDRBoundExcMd.parSize);
              ctx.arch.genRestUnitContext();
            }
            else {
              ctx.arch.genPush(obj, StdTypes.T_PTR);
              ctx.arch.genPush(index, StdTypes.T_INT);
              ctx.arch.genCallConst(ctx.rteDRBoundExcMd, ctx.rteDRBoundExcMd.parSize);
            }
          }
          else ctx.arch.genNativeBoundException();
          ctx.arch.appendInstruction(excCheckDone);
        }
        regType=getRegType(ctx);
        if (regType==StdTypes.T_PTR) size=-ctx.arch.relocBytes;
        else size=TypeRef.getMinSize(regType);
        if (size>0) {
          if (ctx.indirScalars) { //movable indirect scalars
            ctx.arch.genLoadVarVal(addrReg, obj, null, ctx.indirScalarAddrOff, StdTypes.T_PTR);
            ctx.arch.genLoadDerefAddr(addrReg, index, container.instIndirScalarTableSize, size);
          }
          else { //normal scalars
            ctx.arch.genDup(addrReg, obj, StdTypes.T_PTR);
            if (flash) ctx.arch.nextValInFlash();
            ctx.arch.genLoadDerefAddr(addrReg, index, container.instScalarTableSize, size);
          }
        }
        else { //relocs
          ctx.arch.genDup(addrReg, obj, StdTypes.T_PTR);
          if (flash) ctx.arch.nextValInFlash();
          ctx.arch.genLoadDerefAddr(addrReg, index, -(container.instRelocTableEntries+1)*ctx.arch.relocBytes, size);
        }
        ctx.arch.deallocRestoreReg(index, 0, restore);
        break;
      default: //struct inline array with positive limited index
        if (le.typeSpecial<0) {
          compErr(ctx, "ExDeArray.genOutputAddr with unknown typeSpecial");
          return;
        }
        //no break
      case TypeRef.S_STRUCTARRDONTCHECK: //struct inline array with unlimited index
      case TypeRef.S_STRUCTARRNOTSPEC: //struct inline array with unlimited index
        if (baseType>0) size=TypeRef.getMinSize(baseType);
        else if (baseType==TypeRef.T_QID) size=qid.unitDest.instScalarTableSize;
        else {
          compErr(ctx, "ExDeArray.genOutputAddr with invalid struct-deref");
          return;
        }
        if (ind.calcConstantType(ctx)==StdTypes.T_INT) {
          pos=ind.getConstIntValue(ctx);
          if (le.typeSpecial>0 && (pos<0 || pos>=le.typeSpecial)) {
            compErr(ctx, "constant index exceeds bounds of inlined array");
            return;
          }
          pos*=size;
          ctx.arch.genLoadVarAddr(addrReg, obj, null, pos);
        }
        else {
          restore=ctx.arch.prepareFreeReg(obj, addrReg, 0, StdTypes.T_INT);
          index=ctx.arch.allocReg();
          ind.genOutputVal(index, ctx);
          if (ctx.doBoundCheck && le.typeSpecial>0) { //insert check for array index
            excCheckDone=ctx.arch.getUnlinkedInstruction();
            condHnd=ctx.arch.genCompValToConstVal(index, le.typeSpecial, StdTypes.T_INT, Ops.C_BO);
            ctx.arch.genCondJmp(excCheckDone, condHnd);
            if (ctx.runtimeBound) {
              if (ctx.dynaMem) {
                ctx.arch.genSaveUnitContext();
                ctx.arch.genPush(obj, StdTypes.T_PTR);
                ctx.arch.genPush(index, StdTypes.T_INT);
                ctx.arch.genLoadUnitContext(ctx.arch.regClss, runtimeClass.relOff);
                ctx.arch.genCall(ctx.rteDRBoundExcMd.relOff, 0, true,
                    ctx.rteDRBoundExcMd.parSize);
                ctx.arch.genRestUnitContext();
              }
              else {
                ctx.arch.genPush(obj, StdTypes.T_PTR);
                ctx.arch.genPush(index, StdTypes.T_INT);
                ctx.arch.genCallConst(ctx.rteDRBoundExcMd, ctx.rteDRBoundExcMd.parSize);
              }
            }
            else ctx.arch.genNativeBoundException();
            ctx.arch.appendInstruction(excCheckDone);
          }
          ctx.arch.genDup(addrReg, obj, StdTypes.T_PTR);
          ctx.arch.genLoadDerefAddr(addrReg, index, 0, size);
          ctx.arch.deallocRestoreReg(index, 0, restore);
        }
    }
    if (lePtr==0) ctx.arch.deallocRestoreReg(obj, addrReg, objRestore);
  }
  
  public void genOutputVal(int reg, Context ctx) {
    int restore, addr;
    
    restore=ctx.arch.prepareFreeReg(0, 0, reg, StdTypes.T_PTR);
    addr=ctx.arch.allocReg();
    genOutputAddr(addr, ctx);
    if (qid!=null && (qid.unitDest.modifier&Modifier.M_STRUCT)!=0)
      ctx.arch.genDup(reg, addr, StdTypes.T_PTR);
    else ctx.arch.genLoadVarVal(reg, addr, null, 0, getRegType(ctx));
    ctx.arch.deallocRestoreReg(addr, reg, restore);
  }
}
